#pragma once

#include <Eigen/Dense>

#include <dune/common/ftraits.hh>

#include <amdis/Output.hpp>
#include <amdis/linearalgebra/DOFVectorBase.hpp>

namespace AMDiS
{
  /// The basic container that stores a base vector and a corresponding basis
  template <class ValueType>
  class EigenVector
  {
  public:
    /// The type of the elements of the DOFVector
    using value_type = ValueType;

    /// The type of the elements of the DOFVector
    using block_type = ValueType;

    /// The underlying field type
    using field_type = typename Dune::FieldTraits<ValueType>::field_type;

    /// The type of the base vector
    using BaseVector = Eigen::Matrix<ValueType, Eigen::Dynamic, 1>;

    /// The index/size - type
    using size_type  = typename BaseVector::Index;

  public:
    /// Constructor. Constructs new BaseVector.
    EigenVector() = default;

    /// Return the data-vector \ref vector_
    BaseVector const& vector() const
    {
      return vector_;
    }

    /// Return the data-vector \ref vector_
    BaseVector& vector()
    {
      return vector_;
    }

    /// Return the current size of the \ref vector_
    size_type size() const
    {
      return vector_.size();
    }

    /// Resize the \ref vector_ to the size \p s
    void resize(size_type s)
    {
      vector_.resize(s);
    }


    /// Access the entry \p i of the \ref vector with read-access.
    value_type const& operator[](size_type i) const
    {
      test_exit_dbg(i < size(),
        "Index {} out of range [0,{})", i, size());
      return vector_.coeff(i);
    }

    /// Access the entry \p i of the \ref vector with write-access.
    value_type& operator[](size_type i)
    {
      test_exit_dbg(i < size(),
        "Index {} out of range [0,{})", i, size());
      return vector_.coeffRef(i);
    }

    void set(field_type value)
    {
      vector_.setConstant(value);
    }

  private:
    /// The data-vector (can hold a new BaseVector or a pointer to external data
    BaseVector vector_;
  };


  template <class BasisType, class ValueType = double>
  class DOFVector : public DOFVectorBase<BasisType, EigenVector<ValueType>>
  {
    using Super = DOFVectorBase<BasisType, EigenVector<ValueType>>;

  public:
    DOFVector(BasisType const& basis, DataTransferOperation op = DataTransferOperation::INTERPOLATE)
      : Super(basis, op)
    {}

    using Super::operator=;
  };


  /// Constructor a dofvector from given basis and name
  template <class ValueType = double, class Basis>
  DOFVector<Basis, ValueType>
  makeDOFVector(Basis const& basis, DataTransferOperation op = DataTransferOperation::INTERPOLATE)
  {
    return {basis, op};
  }

} // end namespace AMDiS
