#pragma once

#include <string>

// MTL4 headers
#include <boost/numeric/itl/itl.hpp>
#include <boost/numeric/mtl/mtl.hpp>

// AMDiS headers
#include <amdis/linearalgebra/PreconditionerInterface.hpp>
#include <amdis/linearalgebra/RunnerInterface.hpp>
#include <amdis/linearalgebra/SolverInfo.hpp>

#include <amdis/linearalgebra/mtl/ITL_Preconditioner.hpp>

namespace AMDiS
{
  /** \ingroup Solver
   *
   * \brief
   * Wrapper class for different MTL4 itl-solvers. These solvers
   * are parametrized by Matrix and Vector.
   **/
  template <class ITLSolver, class Matrix, class Vector>
  class KrylovRunner
      : public RunnerInterface<Matrix, Vector>
  {
    using Super = RunnerInterface<Matrix, Vector>;
    using PreconBase = typename Super::PreconBase;

  public:
    /// Constructor.
    explicit KrylovRunner(std::string const& prefix)
      : itlSolver_(prefix)
    {
      initPrecon(prefix);

      Parameters::get(prefix + "->max iteration", maxIter_);
      Parameters::get(prefix + "->print cycle", printCycle_);
    }

    /// Implements \ref RunnerInterface::lLeftPrecon(), Returns \ref L_
    std::shared_ptr<PreconBase> leftPrecon() override
    {
      return L_;
    }

    /// Implements \ref RunnerInterface::rightPrecon(), Returns \ref R_
    std::shared_ptr<PreconBase> rightPrecon() override
    {
      return R_;
    }

    /// Set a new left preconditioner \ref L_
    void setLeftPrecon(std::shared_ptr<PreconBase> const& precon)
    {
      L_ = precon;
    }

    /// Set a new right preconditioner \ref R_
    void setRightPrecon(std::shared_ptr<PreconBase> const& precon)
    {
      R_ = precon;
    }

    /// Implementation of \ref RunnerInterface::init()
    void init(Matrix const& A) override
    {
      L_->init(A);
      R_->init(A);
    }

    /// Implementation of \ref RunnerInterface::exit()
    void exit() override
    {
      L_->exit();
      R_->exit();
    }

    /// Implementation of \ref RunnerInterface::solve()
    int solve(Matrix const& A, Vector& x, Vector const& b,
              SolverInfo& solverInfo) override
    {
      test_exit(bool(L_), "There is no left preconditioner");
      test_exit(bool(R_), "There is no right preconditioner");

      auto r = Vector(b);
      if (two_norm(x) != 0)
        r -= A * x;

      // print information about the solution process
      itl::cyclic_iteration<double> iter(r, maxIter_, solverInfo.relTolerance(),
                                                      solverInfo.absTolerance(),
                                                      printCycle_);
      iter.set_quite(solverInfo.info() == 0);
      iter.suppress_resume(solverInfo.info() == 0);

      int error = itlSolver_(A, x, b, *L_, *R_, iter);
      solverInfo.setAbsResidual(iter.resid());
      solverInfo.setRelResidual(iter.relresid());

      return error;
    }

  private:
    /// Create left/right preconditioners from parameters given in the init-file
    void initPrecon(std::string const& prefix)
    {
      // Creator for the left preconditioner
      std::string leftPreconName = "", rightPreconName = "";
      Parameters::get(prefix + "->left precon", leftPreconName);
      Parameters::get(prefix + "->right precon", rightPreconName);

      auto leftCreator
        = CreatorMap<PreconBase>::getCreator(leftPreconName, prefix + "->left precon");
      auto rightCreator
        = CreatorMap<PreconBase>::getCreator(rightPreconName, prefix + "->right precon");

      L_ = leftCreator->create();
      R_ = rightCreator->create();
    }

  private:
    /// The itl solver object that performes the actual solve
    ITLSolver itlSolver_;

    /// Pointer to the left and right preconditioner
    std::shared_ptr<PreconBase> L_, R_;

    /// The maximal number of iterations
    std::size_t maxIter_ = 1000;

    /// Print solver residuals every \ref printCycle 's iteration
    std::size_t printCycle_ = 100;
  };

} // end namespace AMDiS
