#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/utility/LocalBasisCache.hpp>
#include <amdis/Output.hpp>
#include <amdis/common/StaticSize.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct partialtest_partialtrial
    {
      std::size_t comp_test; // i
      std::size_t comp_trial; // j
    };
  }


  /// second-order operator \f$ \langle\partial_i\psi, c\,\partial_j\phi\rangle \f$
  template <class LC, class GridFct>
  class GridFunctionOperator<tag::partialtest_partialtrial, LC, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::partialtest_partialtrial, LC, GridFct>, LC, GridFct>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LC, GridFct>;

    static_assert( Size_v<typename GridFct::Range> == 1, "Expression must be of scalar type." );

  public:
    GridFunctionOperator(tag::partialtest_partialtrial tag, GridFct const& expr)
      : Super(expr, 2)
      , compTest_(tag.comp_test)
      , compTrial_(tag.comp_trial)
    {}

    template <class CG, class RN, class CN, class Mat>
    void getElementMatrix(CG const& contextGeo, RN const& rowNode, CN const& colNode, Mat& elementMatrix)
    {
      static_assert(RN::isLeaf && CN::isLeaf,
        "Operator can be applied to Leaf-Nodes only.");

      auto const& quad = this->getQuadratureRule(contextGeo.type(), rowNode, colNode);
      auto const& rowLocalFE = rowNode.finiteElement();
      auto const& colLocalFE = colNode.finiteElement();
      std::size_t rowSize = rowLocalFE.size();
      std::size_t colSize = colLocalFE.size();

      using RowFieldType = typename NodeQuadCache<RN>::Traits::RangeFieldType;
      using ColFieldType = typename NodeQuadCache<CN>::Traits::RangeFieldType;

      NodeQuadCache<RN> rowCache(rowLocalFE.localBasis());
      NodeQuadCache<CN> colCache(colLocalFE.localBasis());

      auto const& rowGradientsCache = rowCache.evaluateJacobianAtQP(contextGeo, quad);
      auto const& colGradientsCache = colCache.evaluateJacobianAtQP(contextGeo, quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = contextGeo.local(quad[iq].position());

        // The transposed inverse Jacobian of the map from the reference element to the element
        const auto jacobian = contextGeo.geometry().jacobianInverseTransposed(local);
        auto&& jacobian_mat = Dune::MatVec::as_matrix(jacobian);

        // The multiplicative factor in the integral transformation formula
        const auto factor = contextGeo.integrationElement(quad[iq].position()) * quad[iq].weight();
        const auto c = Super::coefficient(local);

        // The gradients of the shape functions on the reference element
        auto const& rowShapeGradients = rowGradientsCache[iq];
        auto const& colShapeGradients = colGradientsCache[iq];

        // Compute the shape function gradients on the real element
        thread_local std::vector<RowFieldType> rowPartial;
        rowPartial.resize(rowShapeGradients.size());
        for (std::size_t i = 0; i < rowPartial.size(); ++i) {
          rowPartial[i] = jacobian_mat[compTest_][0] * rowShapeGradients[i][0][0];
          for (std::size_t j = 1; j < jacobian_mat.M(); ++j)
            rowPartial[i] += jacobian_mat[compTest_][j] * rowShapeGradients[i][0][j];
        }

        thread_local std::vector<ColFieldType> colPartial;
        colPartial.resize(colShapeGradients.size());
        for (std::size_t i = 0; i < colPartial.size(); ++i) {
          colPartial[i] = jacobian_mat[compTrial_][0] * colShapeGradients[i][0][0];
          for (std::size_t j = 1; j < jacobian_mat.M(); ++j)
            colPartial[i] += jacobian_mat[compTrial_][j] * colShapeGradients[i][0][j];
        }

        for (std::size_t j = 0; j < colSize; ++j) {
          const auto local_j = colNode.localIndex(j);
          const auto value = factor * (c * colPartial[j]);
          for (std::size_t i = 0; i < rowSize; ++i) {
            const auto local_i = rowNode.localIndex(i);
            elementMatrix[local_i][local_j] += value * rowPartial[i];
          }
        }
      }

    }

  private:
    std::size_t compTest_;
    std::size_t compTrial_;
  };


  /// Create a second-order term of partial derivatives
  template <class Expr, class... QuadratureArgs>
  auto sot_ij(Expr&& expr, std::size_t comp_test, std::size_t comp_trial, QuadratureArgs&&... args)
  {
    return makeOperator(tag::partialtest_partialtrial{comp_test, comp_trial}, FWD(expr), FWD(args)...);
  }

  /** @} **/

} // end namespace AMDiS
